#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <linux/soundcard.h>

#include "lib/log.h"
#include "sound.h"

//#ifdef NO_16BIT_SOUND
typedef unsigned char Sample;
//#else
//typedef signed short Sample;
//#endif

static int speed;
static fixed volume=FIXED_ONE_HALF;

/* if driver has more than this many frags free, it's time to send more data */
static int slush_fragments;

/* how big a fragment is */
static unsigned int fragment_size;

static int fd=-1;

void init_sound(int s) {
   int bitness=sizeof(Sample)*8,stereority=0;
   audio_buf_info abi;
   int window;

   /* check, are we already inited? */
   if(fd>=0) reset_sound();

   /* try to open dsp */
   fd=open("/dev/dsp",O_WRONLY);
   if(fd<0) {
      logprintf(LOG_ERROR,'S',"init_sound: open failed, errno=%d",errno);
      return;
   };

   /* andconfigure it */
   if(ioctl(fd,SNDCTL_DSP_STEREO,&stereority)) {
      logprintf(LOG_ERROR,'S',
		"init_sound: set %s failed, errno=%d",
		"stereo",errno);
      close(fd);
      fd=-1;
   };
   if(ioctl(fd,SNDCTL_DSP_SETFMT,&bitness)) {
      logprintf(LOG_ERROR,'S',
		"init_sound: set %d-bit failed, errno=%d",
		bitness,errno);
      close(fd);
      fd=-1;
   };
   speed=s;
   if(ioctl(fd,SNDCTL_DSP_SPEED,&speed)) {
      logprintf(LOG_ERROR,'S',
		"init_sound: set %s failed, errno=%d",
		"speed",errno);
      close(fd);
      fd=-1;
   };

   /* figure out window size */
   ioctl(fd,SNDCTL_DSP_GETOSPACE,&abi);
   fragment_size=abi.fragsize;
   if(8*fragment_size<speed) window=speed/fragment_size*2;
   else window=2;
   if(window<2) window=2;
   if(window>=abi.fragstotal) window=abi.fragstotal-1;

   /* and work out slush frags from that */
   slush_fragments=abi.fragstotal-window;

   /* say what we've done */
   logprintf(LOG_INFO,'S',
	     "init_sound: fd=%d speed=%d window=%d (%f seconds) slush=%d",
	     fd,speed,window,
	     (float)(window*fragment_size)/(float)speed,
	     slush_fragments);
};

/* stop playing now, and throw away anything you were about to play */
void purge_sound(void) {
};

/* do all of the above, and shut down the sound device, clean up, etc */
void reset_sound(void) {
   purge_sound();
   if(fd>=0) close(fd);
   fd=-1;
};

void set_sound_volume(fixed vol) {
   volume=vol;
};

void play_sound(const unsigned char *buf,int count,
		SoundBalance bal,int myspeed) {
   logprintf(LOG_DEBUG,'S',"play_sound: %d bytes: %f seconds",
	     count,((float)count)/((float)myspeed));
   if(fd<0) return;
   write(fd,buf,count);
};

void poll_sound() {};








